"""Configure loggers for remote or local execution."""

import logging
import os

# Configure logging for local or remote execution
LOGGING_REMOTE_ENV_VAR = "UNIONAI_SDK_REMOTE_LOGGING_LEVEL"
LOGGING_ENV_VAR = "UNIONAI_SDK_LOGGING_LEVEL"


def _init_global_loggers():
    """Initialize global loggers."""
    logger_configured = _init_logger(LOGGING_ENV_VAR)

    if logger_configured:
        # Logger is already configured, do not need to create another handler
        return

    if "FLYTE_INTERNAL_EXECUTION_ID" in os.environ:
        _init_logger(LOGGING_REMOTE_ENV_VAR)


def _init_logger(env_var: str) -> bool:
    """Return True if logger is configured."""
    logging_level = os.getenv(env_var)
    if logging_level is None:
        return False

    logger = logging.getLogger("unionai")

    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    formatter = logging.Formatter(fmt="[%(name)s] %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel(int(logging_level))

    return True



"""This module contains utils that are useful for development.

Example usage:

```python
import pandas as pd
from flytekit import task, workflow

from unionai._testing import imagespec_with_local_unionai

image = imagespec_with_local_unionai(packages=["pandas"])


@task(container_image=image)
def get_data() -> pd.DataFrame:
    return pd.DataFrame([1, 2, 3])


@task(container_image=image)
def double_df(df: pd.DataFrame) -> pd.DataFrame:
    return 2 * df


@workflow
def main() -> pd.DataFrame:
    df = get_data()
    return double_df(df=df)
```
"""

import os
import sys
from pathlib import Path
from shutil import rmtree
from subprocess import run
from typing import Optional

from flytekit import ImageSpec
from flytekit.core.context_manager import FlyteContextManager

import unionai


def imagespec_with_local_unionai(
    *args, packages: Optional[list] = None, source_root: Optional[str] = None, **kwargs
) -> Optional[ImageSpec]:
    """Creates an imagespec with local changes."""
    state = FlyteContextManager.current_context().execution_state
    if not (state and state.mode is None):
        return None

    unionai_root_path = Path(unionai.__path__[0]).parent

    assert (unionai_root_path / "pyproject.toml").exists(), "pyproject.toml needs to exists in root directory"

    if source_root is None:
        tmp_path = Path("tmp")
        source_root = tmp_path / "source_root"
        if source_root.exists():
            rmtree(source_root)
        source_root.mkdir(parents=True)
    else:
        source_root = Path(source_root)

    vendor_path = source_root / ".vendor"
    vendor_path = vendor_path.absolute()

    run([sys.executable, "-m", "build", "--outdir", vendor_path, "--wheel"], check=True, cwd=unionai_root_path)
    wheels = list(vendor_path.iterdir())
    assert len(wheels) == 1, f"There should be only one wheel in {vendor_path}"
    name = wheels[0].name

    packages = packages or []
    packages.append(f".vendor/{name}")

    source_root_absolute = source_root.absolute()
    return ImageSpec(*args, packages=packages, source_root=os.fspath(source_root_absolute), **kwargs)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from unionai.internal.artifacts import artifacts_pb2 as flyteidl_dot_artifacts_dot_artifacts__pb2


class ArtifactRegistryStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.CreateArtifact = channel.unary_unary(
                '/flyteidl.artifact.ArtifactRegistry/CreateArtifact',
                request_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.CreateArtifactRequest.SerializeToString,
                response_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.CreateArtifactResponse.FromString,
                )
        self.GetArtifact = channel.unary_unary(
                '/flyteidl.artifact.ArtifactRegistry/GetArtifact',
                request_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.GetArtifactRequest.SerializeToString,
                response_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.GetArtifactResponse.FromString,
                )
        self.SearchArtifacts = channel.unary_unary(
                '/flyteidl.artifact.ArtifactRegistry/SearchArtifacts',
                request_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.SearchArtifactsRequest.SerializeToString,
                response_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.SearchArtifactsResponse.FromString,
                )
        self.CreateTrigger = channel.unary_unary(
                '/flyteidl.artifact.ArtifactRegistry/CreateTrigger',
                request_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.CreateTriggerRequest.SerializeToString,
                response_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.CreateTriggerResponse.FromString,
                )
        self.ActivateTrigger = channel.unary_unary(
                '/flyteidl.artifact.ArtifactRegistry/ActivateTrigger',
                request_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.ActivateTriggerRequest.SerializeToString,
                response_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.ActivateTriggerResponse.FromString,
                )
        self.DeactivateTrigger = channel.unary_unary(
                '/flyteidl.artifact.ArtifactRegistry/DeactivateTrigger',
                request_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.DeactivateTriggerRequest.SerializeToString,
                response_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.DeactivateTriggerResponse.FromString,
                )
        self.DeactivateAllTriggers = channel.unary_unary(
                '/flyteidl.artifact.ArtifactRegistry/DeactivateAllTriggers',
                request_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.DeactivateAllTriggersRequest.SerializeToString,
                response_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.DeactivateAllTriggersResponse.FromString,
                )
        self.GetCard = channel.unary_unary(
                '/flyteidl.artifact.ArtifactRegistry/GetCard',
                request_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.GetCardRequest.SerializeToString,
                response_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.GetCardResponse.FromString,
                )
        self.AddTag = channel.unary_unary(
                '/flyteidl.artifact.ArtifactRegistry/AddTag',
                request_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.AddTagRequest.SerializeToString,
                response_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.AddTagResponse.FromString,
                )
        self.RegisterProducer = channel.unary_unary(
                '/flyteidl.artifact.ArtifactRegistry/RegisterProducer',
                request_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.RegisterProducerRequest.SerializeToString,
                response_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.RegisterResponse.FromString,
                )
        self.RegisterConsumer = channel.unary_unary(
                '/flyteidl.artifact.ArtifactRegistry/RegisterConsumer',
                request_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.RegisterConsumerRequest.SerializeToString,
                response_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.RegisterResponse.FromString,
                )
        self.SetExecutionInputs = channel.unary_unary(
                '/flyteidl.artifact.ArtifactRegistry/SetExecutionInputs',
                request_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.ExecutionInputsRequest.SerializeToString,
                response_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.ExecutionInputsResponse.FromString,
                )
        self.FindByWorkflowExec = channel.unary_unary(
                '/flyteidl.artifact.ArtifactRegistry/FindByWorkflowExec',
                request_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.FindByWorkflowExecRequest.SerializeToString,
                response_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.SearchArtifactsResponse.FromString,
                )
        self.ListUsage = channel.unary_unary(
                '/flyteidl.artifact.ArtifactRegistry/ListUsage',
                request_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.ListUsageRequest.SerializeToString,
                response_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.ListUsageResponse.FromString,
                )
        self.GetTriggeringArtifacts = channel.unary_unary(
                '/flyteidl.artifact.ArtifactRegistry/GetTriggeringArtifacts',
                request_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.GetTriggeringArtifactsRequest.SerializeToString,
                response_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.GetTriggeringArtifactsResponse.FromString,
                )
        self.GetTriggeredExecutionsByArtifact = channel.unary_unary(
                '/flyteidl.artifact.ArtifactRegistry/GetTriggeredExecutionsByArtifact',
                request_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.GetTriggeredExecutionsByArtifactRequest.SerializeToString,
                response_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.GetTriggeredExecutionsByArtifactResponse.FromString,
                )


class ArtifactRegistryServicer(object):
    """Missing associated documentation comment in .proto file."""

    def CreateArtifact(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetArtifact(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SearchArtifacts(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateTrigger(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ActivateTrigger(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeactivateTrigger(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeactivateAllTriggers(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetCard(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def AddTag(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RegisterProducer(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RegisterConsumer(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetExecutionInputs(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FindByWorkflowExec(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ListUsage(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTriggeringArtifacts(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTriggeredExecutionsByArtifact(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ArtifactRegistryServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'CreateArtifact': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateArtifact,
                    request_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.CreateArtifactRequest.FromString,
                    response_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.CreateArtifactResponse.SerializeToString,
            ),
            'GetArtifact': grpc.unary_unary_rpc_method_handler(
                    servicer.GetArtifact,
                    request_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.GetArtifactRequest.FromString,
                    response_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.GetArtifactResponse.SerializeToString,
            ),
            'SearchArtifacts': grpc.unary_unary_rpc_method_handler(
                    servicer.SearchArtifacts,
                    request_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.SearchArtifactsRequest.FromString,
                    response_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.SearchArtifactsResponse.SerializeToString,
            ),
            'CreateTrigger': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateTrigger,
                    request_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.CreateTriggerRequest.FromString,
                    response_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.CreateTriggerResponse.SerializeToString,
            ),
            'ActivateTrigger': grpc.unary_unary_rpc_method_handler(
                    servicer.ActivateTrigger,
                    request_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.ActivateTriggerRequest.FromString,
                    response_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.ActivateTriggerResponse.SerializeToString,
            ),
            'DeactivateTrigger': grpc.unary_unary_rpc_method_handler(
                    servicer.DeactivateTrigger,
                    request_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.DeactivateTriggerRequest.FromString,
                    response_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.DeactivateTriggerResponse.SerializeToString,
            ),
            'DeactivateAllTriggers': grpc.unary_unary_rpc_method_handler(
                    servicer.DeactivateAllTriggers,
                    request_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.DeactivateAllTriggersRequest.FromString,
                    response_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.DeactivateAllTriggersResponse.SerializeToString,
            ),
            'GetCard': grpc.unary_unary_rpc_method_handler(
                    servicer.GetCard,
                    request_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.GetCardRequest.FromString,
                    response_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.GetCardResponse.SerializeToString,
            ),
            'AddTag': grpc.unary_unary_rpc_method_handler(
                    servicer.AddTag,
                    request_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.AddTagRequest.FromString,
                    response_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.AddTagResponse.SerializeToString,
            ),
            'RegisterProducer': grpc.unary_unary_rpc_method_handler(
                    servicer.RegisterProducer,
                    request_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.RegisterProducerRequest.FromString,
                    response_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.RegisterResponse.SerializeToString,
            ),
            'RegisterConsumer': grpc.unary_unary_rpc_method_handler(
                    servicer.RegisterConsumer,
                    request_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.RegisterConsumerRequest.FromString,
                    response_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.RegisterResponse.SerializeToString,
            ),
            'SetExecutionInputs': grpc.unary_unary_rpc_method_handler(
                    servicer.SetExecutionInputs,
                    request_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.ExecutionInputsRequest.FromString,
                    response_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.ExecutionInputsResponse.SerializeToString,
            ),
            'FindByWorkflowExec': grpc.unary_unary_rpc_method_handler(
                    servicer.FindByWorkflowExec,
                    request_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.FindByWorkflowExecRequest.FromString,
                    response_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.SearchArtifactsResponse.SerializeToString,
            ),
            'ListUsage': grpc.unary_unary_rpc_method_handler(
                    servicer.ListUsage,
                    request_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.ListUsageRequest.FromString,
                    response_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.ListUsageResponse.SerializeToString,
            ),
            'GetTriggeringArtifacts': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTriggeringArtifacts,
                    request_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.GetTriggeringArtifactsRequest.FromString,
                    response_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.GetTriggeringArtifactsResponse.SerializeToString,
            ),
            'GetTriggeredExecutionsByArtifact': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTriggeredExecutionsByArtifact,
                    request_deserializer=flyteidl_dot_artifacts_dot_artifacts__pb2.GetTriggeredExecutionsByArtifactRequest.FromString,
                    response_serializer=flyteidl_dot_artifacts_dot_artifacts__pb2.GetTriggeredExecutionsByArtifactResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'flyteidl.artifact.ArtifactRegistry', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ArtifactRegistry(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def CreateArtifact(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/flyteidl.artifact.ArtifactRegistry/CreateArtifact',
            flyteidl_dot_artifacts_dot_artifacts__pb2.CreateArtifactRequest.SerializeToString,
            flyteidl_dot_artifacts_dot_artifacts__pb2.CreateArtifactResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetArtifact(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/flyteidl.artifact.ArtifactRegistry/GetArtifact',
            flyteidl_dot_artifacts_dot_artifacts__pb2.GetArtifactRequest.SerializeToString,
            flyteidl_dot_artifacts_dot_artifacts__pb2.GetArtifactResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SearchArtifacts(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/flyteidl.artifact.ArtifactRegistry/SearchArtifacts',
            flyteidl_dot_artifacts_dot_artifacts__pb2.SearchArtifactsRequest.SerializeToString,
            flyteidl_dot_artifacts_dot_artifacts__pb2.SearchArtifactsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CreateTrigger(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/flyteidl.artifact.ArtifactRegistry/CreateTrigger',
            flyteidl_dot_artifacts_dot_artifacts__pb2.CreateTriggerRequest.SerializeToString,
            flyteidl_dot_artifacts_dot_artifacts__pb2.CreateTriggerResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ActivateTrigger(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/flyteidl.artifact.ArtifactRegistry/ActivateTrigger',
            flyteidl_dot_artifacts_dot_artifacts__pb2.ActivateTriggerRequest.SerializeToString,
            flyteidl_dot_artifacts_dot_artifacts__pb2.ActivateTriggerResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeactivateTrigger(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/flyteidl.artifact.ArtifactRegistry/DeactivateTrigger',
            flyteidl_dot_artifacts_dot_artifacts__pb2.DeactivateTriggerRequest.SerializeToString,
            flyteidl_dot_artifacts_dot_artifacts__pb2.DeactivateTriggerResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def DeactivateAllTriggers(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/flyteidl.artifact.ArtifactRegistry/DeactivateAllTriggers',
            flyteidl_dot_artifacts_dot_artifacts__pb2.DeactivateAllTriggersRequest.SerializeToString,
            flyteidl_dot_artifacts_dot_artifacts__pb2.DeactivateAllTriggersResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetCard(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/flyteidl.artifact.ArtifactRegistry/GetCard',
            flyteidl_dot_artifacts_dot_artifacts__pb2.GetCardRequest.SerializeToString,
            flyteidl_dot_artifacts_dot_artifacts__pb2.GetCardResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def AddTag(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/flyteidl.artifact.ArtifactRegistry/AddTag',
            flyteidl_dot_artifacts_dot_artifacts__pb2.AddTagRequest.SerializeToString,
            flyteidl_dot_artifacts_dot_artifacts__pb2.AddTagResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RegisterProducer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/flyteidl.artifact.ArtifactRegistry/RegisterProducer',
            flyteidl_dot_artifacts_dot_artifacts__pb2.RegisterProducerRequest.SerializeToString,
            flyteidl_dot_artifacts_dot_artifacts__pb2.RegisterResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RegisterConsumer(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/flyteidl.artifact.ArtifactRegistry/RegisterConsumer',
            flyteidl_dot_artifacts_dot_artifacts__pb2.RegisterConsumerRequest.SerializeToString,
            flyteidl_dot_artifacts_dot_artifacts__pb2.RegisterResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetExecutionInputs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/flyteidl.artifact.ArtifactRegistry/SetExecutionInputs',
            flyteidl_dot_artifacts_dot_artifacts__pb2.ExecutionInputsRequest.SerializeToString,
            flyteidl_dot_artifacts_dot_artifacts__pb2.ExecutionInputsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def FindByWorkflowExec(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/flyteidl.artifact.ArtifactRegistry/FindByWorkflowExec',
            flyteidl_dot_artifacts_dot_artifacts__pb2.FindByWorkflowExecRequest.SerializeToString,
            flyteidl_dot_artifacts_dot_artifacts__pb2.SearchArtifactsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def ListUsage(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/flyteidl.artifact.ArtifactRegistry/ListUsage',
            flyteidl_dot_artifacts_dot_artifacts__pb2.ListUsageRequest.SerializeToString,
            flyteidl_dot_artifacts_dot_artifacts__pb2.ListUsageResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetTriggeringArtifacts(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/flyteidl.artifact.ArtifactRegistry/GetTriggeringArtifacts',
            flyteidl_dot_artifacts_dot_artifacts__pb2.GetTriggeringArtifactsRequest.SerializeToString,
            flyteidl_dot_artifacts_dot_artifacts__pb2.GetTriggeringArtifactsResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def GetTriggeredExecutionsByArtifact(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/flyteidl.artifact.ArtifactRegistry/GetTriggeredExecutionsByArtifact',
            flyteidl_dot_artifacts_dot_artifacts__pb2.GetTriggeredExecutionsByArtifactRequest.SerializeToString,
            flyteidl_dot_artifacts_dot_artifacts__pb2.GetTriggeredExecutionsByArtifactResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



"""Module for fsspec implementations."""
from unionai.filesystems._unionfs import AsyncUnionFS
from unionai.filesystems._unionmetafs import AsyncUnionMetaFS

__all__ = ["AsyncUnionFS", "AsyncUnionMetaFS"]



"""Module to track usage and errors."""

import logging
from functools import partial
from os import getenv
from pathlib import Path
from typing import Any

import click
import flytekit
from flytekit.clis.sdk_in_container.utils import get_level_from_cli_verbosity, pretty_print_exception
from flytekit.exceptions.base import FlyteException
from flytekit.loggers import upgrade_to_rich_logging
from importlib_metadata import version
from rich_click.rich_command import RichGroup

import unionai

SENTRY_DSN = "https://8a7049a7fe42d7fe7b12e1b0d2171d21@o4507249423810560.ingest.us.sentry.io/4507249829281792"
SENTRY_DEBUG_DSN = "https://f9933bb58495729cffc5f6b97ea8502e@o4507249423810560.ingest.us.sentry.io/4507249428201472"
UNIONAI_USAGE_STATS_ENABLED_ENV = "UNIONAI_USAGE_STATS_ENABLED"
UNIONAI_DEBUG_USAGE_STATS_ENV = "UNIONAI_DEBUG_USAGE_STATS"


# Re-implementation of ErrorHandlingCommand.invoke to send error to sentry first
# Details: ErrorHandlingCommand.invoke captures the error and calls exit(1). This means
# we need to override `ErrorHandlingCommand.invoke` and capture the exceptions before
# `exit(1)`
# https://github.com/flyteorg/flytekit/blob/db2ff9e4e14bdcec2c96388af33f0a3357459fe0/flytekit/clis/sdk_in_container/utils.py#L123-L141
def custom_invoke(ctx: click.Context, self) -> Any:
    verbose = ctx.params["verbose"]
    log_level = get_level_from_cli_verbosity(verbose)
    upgrade_to_rich_logging(log_level=log_level)
    try:
        return RichGroup.invoke(self, ctx)
    except Exception as e:
        import sentry_sdk

        sentry_sdk.capture_exception(e)

        if verbose > 0:
            click.secho("Verbose mode on")
            if isinstance(e, FlyteException):
                raise e.with_traceback(None)
            raise e
        pretty_print_exception(e)
        exit(1)


# Only show exceptions caused by unionai and flytekit.clients
PACKAGE_PATHS_TO_TRACK = unionai.__path__ + flytekit.clients.__path__


def _sentry_exit_callback(pending, timeout):
    """Silence 'Sentry is attempted to send... message"""
    pass


def _filter_exceptions(event, hint):
    """Filter out exceptions that are not from the unionai SDK."""
    try:
        source = event.get("exception") or event.get("threads")

        # Ignore events without a stack trace
        if source is None:
            return None

        last_frame = source["values"][0]["stacktrace"]["frames"][-1]
        exc_origin_path = last_frame["abs_path"]

        if not any(exc_origin_path.startswith(p) for p in PACKAGE_PATHS_TO_TRACK):
            return None
    except Exception:
        return None

    return event


def _configure_tracking_cli(main: click.Group) -> click.Group:
    """Configures the the unionai to catch exceptions."""
    main.invoke = partial(custom_invoke, self=main)
    _init_tracking()
    return main


def _init_tracking():
    """Initializes tracking."""
    # Check if this is a development environment
    unionai_git_path = Path(unionai.__path__[0]).parent / ".git"
    in_dev_env = unionai_git_path.exists()

    # If UNIONAI_DEBUG_SENTRY_ENV != "0", then enable user stats for development.
    # If UNIONAI_DEBUG_SENTRY_ENV == "0", then disable user stats for development. (Default)
    enable_tracking_for_dev = getenv(UNIONAI_DEBUG_USAGE_STATS_ENV, "0") != "0" and in_dev_env

    # If UNIONAI_USAGE_STATS_ENABLED_ENV != "0", then enable user stats collection. (Default)
    # If UNIONAI_USAGE_STATS_ENABLED_ENV == "0", then disable user stats collection.
    enable_usage_stats = getenv(UNIONAI_USAGE_STATS_ENABLED_ENV, "1") != "0" and enable_tracking_for_dev

    if not enable_usage_stats:
        return

    import sentry_sdk
    from sentry_sdk.integrations.atexit import AtexitIntegration
    from sentry_sdk.integrations.logging import LoggingIntegration

    sentry_logging = LoggingIntegration(
        level=logging.DEBUG,  # Capture debug and above as breadcrumbs
        event_level=logging.ERROR,  # Send errors as events
    )

    if enable_tracking_for_dev:
        # Use debug DSN for development, so we do not put these errors into the production tracker
        dsn = SENTRY_DEBUG_DSN
        release = "dev"
        environment = "development"
    else:
        dsn = SENTRY_DSN
        release = version("unionai")
        environment = "production"

    sentry_sdk.init(
        dsn=dsn,
        auto_enabling_integrations=False,
        before_send=_filter_exceptions,
        integrations=[AtexitIntegration(callback=_sentry_exit_callback), sentry_logging],
        enable_tracing=False,
        release=release,
        environment=environment,
        debug=enable_tracking_for_dev,
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: identity/enums.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14identity/enums.proto\x12\x11\x63loudidl.identity*2\n\rResponseTypes\x12\x08\n\x04\x43ODE\x10\x00\x12\t\n\x05TOKEN\x10\x01\x12\x0c\n\x08ID_TOKEN\x10\x02*&\n\x0f\x41pplicationType\x12\x07\n\x03WEB\x10\x00\x12\n\n\x06NATIVE\x10\x01*|\n\nGrantTypes\x12\x16\n\x12\x41UTHORIZATION_CODE\x10\x00\x12\x0c\n\x08IMPLICIT\x10\x01\x12\x0c\n\x08PASSWORD\x10\x02\x12\x16\n\x12\x43LIENT_CREDENTIALS\x10\x03\x12\x11\n\rREFRESH_TOKEN\x10\x04\x12\x0f\n\x0b\x44\x45VICE_CODE\x10\x05*T\n\x17TokenEndpointAuthMethod\x12\x08\n\x04NONE\x10\x00\x12\x16\n\x12\x43LIENT_SECRET_POST\x10\x01\x12\x17\n\x13\x43LIENT_SECRET_BASIC\x10\x02*a\n\rConsentMethod\x12\x17\n\x13\x43ONSENT_METHOD_NONE\x10\x00\x12\x1a\n\x16\x43ONSENT_METHOD_TRUSTED\x10\x01\x12\x1b\n\x17\x43ONSENT_METHOD_REQUIRED\x10\x02\x42\xb7\x01\n\x15\x63om.cloudidl.identityB\nEnumsProtoH\x02P\x01Z+github.com/unionai/cloud/gen/pb-go/identity\xa2\x02\x03\x43IX\xaa\x02\x11\x43loudidl.Identity\xca\x02\x11\x43loudidl\\Identity\xe2\x02\x1d\x43loudidl\\Identity\\GPBMetadata\xea\x02\x12\x43loudidl::Identityb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'identity.enums_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\025com.cloudidl.identityB\nEnumsProtoH\002P\001Z+github.com/unionai/cloud/gen/pb-go/identity\242\002\003CIX\252\002\021Cloudidl.Identity\312\002\021Cloudidl\\Identity\342\002\035Cloudidl\\Identity\\GPBMetadata\352\002\022Cloudidl::Identity'
  _globals['_RESPONSETYPES']._serialized_start=43
  _globals['_RESPONSETYPES']._serialized_end=93
  _globals['_APPLICATIONTYPE']._serialized_start=95
  _globals['_APPLICATIONTYPE']._serialized_end=133
  _globals['_GRANTTYPES']._serialized_start=135
  _globals['_GRANTTYPES']._serialized_end=259
  _globals['_TOKENENDPOINTAUTHMETHOD']._serialized_start=261
  _globals['_TOKENENDPOINTAUTHMETHOD']._serialized_end=345
  _globals['_CONSENTMETHOD']._serialized_start=347
  _globals['_CONSENTMETHOD']._serialized_end=444
# @@protoc_insertion_point(module_scope)



from __future__ import annotations

import datetime as _datetime
import hashlib
import logging
from dataclasses import asdict, dataclass
from functools import partial
from typing import Any, Callable, Dict, Iterable, List, Optional, Union

from flytekit.configuration import SerializationSettings
from flytekit.core import launch_plan as _annotated_launchplan
from flytekit.core import workflow as _annotated_workflow
from flytekit.core.base_task import TaskResolverMixin
from flytekit.core.python_function_task import PythonFunctionTask
from flytekit.core.resources import Resources
from flytekit.core.task import TaskPlugins
from flytekit.core.task import task as flytekit_task_decorator
from flytekit.image_spec import ImageSpec
from flytekit.models.documentation import Documentation
from flytekit.models.security import Secret

logger = logging.getLogger(__name__)


@dataclass
class ActorEnvironment:
    """
    ActorEnvironment class.

    :param container_image: The container image to use for the task. Set to
        default image if none provided.
    :param backlog_length: The number of tasks to keep in the worker queue on
        the backend. Setting `backlog_length` ensures that the worker
        executing actor tasks immediately executes the next task after
        completing the previous one instead of waiting for the scheduler to
        complete other operations before scheduling the next task.
    :param parallelism: The number of tasks that can execute in parallel,
        per worker.
    :param replica_count: The number of workers to provision that are able to
        accept tasks.
    :param ttl_seconds: How long to keep the Actor alive while no tasks are
        being run.
    :param environment: Environment variables as key, value pairs in a Python
        dictionary.
    :param requests: Compute resource requests per task.
    :param limits: Compute resource limits.
    :param secret_requests: Keys (ideally descriptive) that can identify the
        secrets supplied at runtime.
    """

    # I think we should add a user-defined name field here to help with entropy, but there are concerns also
    # These fields here are part of the actor config.
    container_image: Optional[Union[str, ImageSpec]] = None
    backlog_length: Optional[int] = None
    parallelism: Optional[int] = None
    replica_count: Optional[int] = None
    ttl_seconds: Optional[int] = None
    environment: Optional[Dict[str, str]] = None
    requests: Optional[Resources] = None
    limits: Optional[Resources] = None
    secret_requests: Optional[List[Secret]] = None

    def __call__(
        self,
        _task_function=None,
        cache: bool = False,
        cache_serialize: bool = False,
        cache_version: str = "",
        retries: int = 0,
        timeout: Union[_datetime.timedelta, int] = 0,
        node_dependency_hints: Optional[
            Iterable[
                Union[
                    PythonFunctionTask,
                    _annotated_launchplan.LaunchPlan,
                    _annotated_workflow.WorkflowBase,
                ]
            ]
        ] = None,
        task_resolver: Optional[TaskResolverMixin] = None,
        docs: Optional[Documentation] = None,
        **kwargs,
    ):
        wrapper = partial(
            flytekit_task_decorator,
            task_config=self,
            cache=cache,
            cache_serialize=cache_serialize,
            cache_version=cache_version,
            retries=retries,
            timeout=timeout,
            node_dependency_hints=node_dependency_hints,
            task_resolver=task_resolver,
            docs=docs,
            container_image=self.container_image,
            environment=self.environment,
            requests=self.requests,
            limits=self.limits,
            secret_requests=self.secret_requests,
            **kwargs,
        )

        if _task_function:
            return wrapper(_task_function=_task_function)
        return wrapper

    @property
    def environment_id(self) -> str:
        # TODO: Add proper entropy
        all_fields = "".join([str(getattr(self, f)) for f in asdict(self).keys()])
        hash_object = hashlib.md5(all_fields.encode("utf-8"))
        hex_digest = hash_object.hexdigest()
        logger.debug(f"Computing actor hash with string {all_fields} to {hex_digest[:15]}")
        return hex_digest[:15]

    @property
    def container_image_name(self) -> str:
        if isinstance(self.container_image, ImageSpec):
            return self.container_image.image_name()
        return self.container_image


class ActorTask(PythonFunctionTask[ActorEnvironment]):
    _ACTOR_TASK_TYPE = "fast-task"

    def __init__(self, task_config: ActorEnvironment, task_function: Callable, **kwargs):
        super(ActorTask, self).__init__(
            task_config=task_config,
            task_type=self._ACTOR_TASK_TYPE,
            task_function=task_function,
            **kwargs,
        )

    def get_custom(self, settings: SerializationSettings) -> Optional[Dict[str, Any]]:
        """
        Serialize the `ActorTask` config into a dict.

        :param settings: Current serialization settings
        :return: Dictionary representation of the dask task config.
        """
        return {
            "id": self.task_config.environment_id,
            "type": self._ACTOR_TASK_TYPE,
            "spec": {
                "container_image": self.task_config.container_image_name,
                "backlog_length": self.task_config.backlog_length,
                "parallelism": self.task_config.parallelism,
                "replica_count": self.task_config.replica_count,
                "ttl_seconds": self.task_config.ttl_seconds,
            },
        }


TaskPlugins.register_pythontask_plugin(ActorEnvironment, ActorTask)
